from MuVisitor import MuVisitor
from MuParser import *
from math import *
#https://github.com/bkiers/Mu/blob/master/src/main/java/mu/EvalVisitor.java


from CFG import *
from antlr4.tree.Trees import Trees 

class MyMuCodeGenVisitor(MuVisitor):
	
	def __init__(self,d,s1,s2,parser):
		self._outputname=s1
		self._outputnamedot=s2
		self._parser=parser
		self._debug=d
		self._memory=dict()
	        #3-address code generation inside CFG.
                self._cfg=CFG(BlockPROG())

	def printRegisterMap(self):
                #TODO
                #Modify after final code gen to say where
                #to find the variables (registers or memory)
		print("--variables to register map--")
		for keys,values in self._memory.items():
			print(keys+'-->'+values)

	def visitParExpr(self,ctx):
		return self.visit(ctx.expr())

	def visitNumberAtom(self,ctx):
		s = ctx.getText();
		try:
			val=int(s)
			#this is only valid if val is representable on 5 bits
			dr = self._cfg.newtmp()
			self._cfg.append(BlockAND(dr,dr,0))
			self._cfg.append(BlockADD(dr,dr,val))
			#self._prog.addInstructionAND(dr,dr,0)
			#self._prog.addInstructionADD(dr,dr,val)
			return dr
		except ValueError:
			raise Exception("Not Yet Implemented")
		
	def visitBooleanAtom(self, ctx):
		#true is 1 false is 0
		b= ctx.getText();
		dr = self._cfg.newtmp()
		#self._prog.addInstructionAND(dr,dr,0)
		self._cfg.append(BlockAND(dr,dr,0))
		if (b=='true'):
			#self._prog.addInstructionADD(dr,dr,1)
			self._cfg.append(BlockADD(dr,dr,1))
		return dr
	
	def visitIdAtom(self,ctx):
		try:
			regval=self._memory[ctx.getText()] #get the register associated to id
			return regval;
		except: #todo
			pass

	def visitStringAtom(self,ctx):
		raise Exception("Not Yet Implemented")

	def visitNilAtom(self,ctx): #todo: see if ok
		raise Exception("Not Yet Implemented")

	#now visit expr
	
	def visitAtomExpr(self,ctx):
		return self.visit(ctx.atom())

	def visitAdditiveExpr(self,ctx): #plus or minus.
		tmpl=self.visit(ctx.expr(0))
		tmpr=self.visit(ctx.expr(1))
		dr = self._cfg.newtmp()

		if (ctx.myop.type==MuParser.PLUS):
			#self._prog.addInstructionADD(dr,tmpl,tmpr)
			self._cfg.append(BlockADD(dr,tmpl,tmpr))
		else: #a +  (-b) 
			negtmpr= self._cfg.newtmp()
			#self._prog.addInstructionNOT(negtmpr,tmpr)
			#self._prog.addInstructionADD(negtmpr,negtmpr,1)
			#self._prog.addInstructionADD(dr,tmpl,negtmpr)
			self._cfg.append(BlockNOT(negtmpr,tmpr))
			self._cfg.append(BlockADD(negtmpr,negtmpr,1))
			self._cfg.append(BlockADD(dr,tmpl,negtmpr))
		return dr

	def visitOrExpr(self,ctx):
		raise Exception("Not Yet Implemented")

	def visitAndExpr(self,ctx):
		raise Exception("Not Yet Implemented")

	def visitEqualityExpr(self,ctx):
		raise Exception("Not Yet Implemented")
		
	def visitRelationalExpr(self,ctx):
		if self._debug:
			print("relational expression:")
			print(Trees.toStringTree(ctx,None,self._parser))
		
		tmpl=self.visit(ctx.expr(0))
		tmpr=self.visit(ctx.expr(1))
		subdr = self._cfg.newtmp()  #e1-e2
		negtmpr= self._cfg.newtmp()
		relexpdr = self._cfg.newtmp() #testdestinationregister
		self._cfg.append(BlockNOT(negtmpr,tmpr))
		self._cfg.append(BlockADD(negtmpr,negtmpr,1))
		self._cfg.append(BlockADD(subdr,tmpl,negtmpr))
		#self._prog.addInstructionNOT(negtmpr,tmpr)
		#self._prog.addInstructionADD(negtmpr,negtmpr,1)
		#self._prog.addInstructionADD(subdr,tmpl,negtmpr)
		
		labeltrue,labelfalse,labelend=self._cfg.newlabelCond()
		bLabeltrue = BlockLABEL(labeltrue)
		bLabelfalse = BlockLABEL(labelfalse)

		# Splitting
		if (ctx.myop.type==MuParser.GT):
			#self._prog.addInstructionBR("zn",labelneg)
			bzn = BlockBR("zn",labelfalse)
			bzn.append(bLabeltrue)
			bzn.append(bLabelfalse)
			self._cfg.append(bzn)
		elif (ctx.myop.type==MuParser.LT):
			#self._prog.addInstructionBR("zp",labelneg)
			bzp = BlockBR("zp",labelfalse)
			bzp.append(bLabeltrue)
			bzp.append(bLabelfalse)
			self._cfg.append(bzp)
		elif (ctx.myop.type==MuParser.GTEQ):
			#self._prog.addInstructionBR("n",labelneg)
			bn = BlockBR("n",labelfalse)
			bn.append(bLabeltrue)
			bn.append(bLabelfalse)
			self._cfg.append(bn)
		else: #lteq
			if self._debug:
				print "<="
			#self._prog.addInstructionBR("p",labelneg)
			bp = BlockBR("p",labelfalse)
			bp.append(bLabeltrue)
			bp.append(bLabelfalse)
			self._cfg.append(bp)
		
		#self._prog.addInstructionAND(relexpdr,relexpdr,0) #true=1
		#self._prog.addInstructionADD(relexpdr,relexpdr,1)
		#self._prog.addInstructionGOTO(labelend)
		#self._prog.addLabel(labelneg)
		#self._prog.addInstructionAND(relexpdr,relexpdr,0) #false=0
		#self._prog.addLabel(labelend)

		# Branch True
		self._cfg.setEnd(bLabeltrue)
		self._cfg.append(BlockAND(relexpdr,relexpdr,0)) #true=1
		self._cfg.append(BlockADD(relexpdr,relexpdr,1))
		endTrue = self._cfg.append(BlockGOTO(labelend))

		# Branch False
		self._cfg.setEnd(bLabelfalse)
		endFalse = self._cfg.append(BlockAND(relexpdr,relexpdr,0))

		# Merging
		bLabelend = BlockLABEL(labelend)
		endTrue.append(bLabelend)
		endFalse.append(bLabelend)

		self._cfg.setEnd(bLabelend)

		return relexpdr

	def visitMultiplicativeExpr(self,ctx):
		raise Exception("Not Yet Implemented")
	
	def visitNotExpr(self,ctx):
		reg=self.visit(ctx.expr())
		dr=self._cfg.newtmp()
		#self._prog.addInstructionNOT(dr,reg)
		self._cfg.append(BlockNOT(dr,reg))
		return dr

	def visitUnaryMinusExpr(selt,ctx):
		raise Exception("Not Yet Implemented")

	def visitPowEpr(self,ctx):
		raise Exception("Not Yet Implemented")

#MAIN VISITOR : returns the CFG.
	def visitProgRule(self,ctx):
		self.visit(ctx.block())
		self.printRegisterMap()
		return self._cfg

#statements
	def visitAssignStat(self,ctx):
		if self._debug:
			print("assign statement, rightexpression is:")
			print(Trees.toStringTree(ctx.expr(),None,self._parser))

		reg4expr=self.visit(ctx.expr())
		name=ctx.ID().getText()
                #this code generation has been modify to solve the previous bug with alias names.
		if name not in self._memory:
			self._memory[name] = self._cfg.newtmp()

                self._cfg.append(BlockADD(self._memory[name],reg4expr,0))	
		

		
        def visitCondBlock(self,ctx):
                if self._debug:
                        print("condblockstatement, condition is:")
                        print(Trees.toStringTree(ctx.expr(),None,self._parser))
                        print("and block is:")
                        print(Trees.toStringTree(ctx.stat_block(),None,self._parser))            
                #exprbool, then true branch goes to the last label (global var).
                #false branch generated after true branch.
                        
                reg=self.visit(ctx.expr()) #generate the code for the test
                
                labeltrue,labelfalse,_=self._cfg.newlabelIf()
                labelveryend = self._lastlabel     #saved by the calling "if"

                #true branch
                #self._prog.addInstructionBR("z",labelfalse) 
    		endCond = self._cfg.append(BlockBR("z",labelfalse))
		endTrue = endCond.append(BlockLABEL(labeltrue))
		endFalse = endCond.append(BlockLABEL(labelfalse))            
                
   		self._cfg.setEnd(endTrue)
		self.visit(ctx.stat_block())
		endTrue = self._cfg.append(BlockGOTO(labelveryend))
                return endTrue,endFalse


        
        def visitIfStat(self,ctx):
                if self._debug:
                        print("if statement")

                L=[]
                _,_,verylabelend=self._cfg.newlabelIf()
                self._lastlabel=verylabelend
                
                for cb in ctx.condition_block():
                        endT,endF=self.visit(cb) 
                        L.append(endT)
                        self._cfg.setEnd(endF) #new end : false branch
                        
                if ctx.stat_block() is not None:
                        if self._debug:
                                print("else  ")
                        self.visit(ctx.stat_block())  #else branch code

                #chain all end of true branches with the global end.
                bLabelend = BlockLABEL(verylabelend)
                for endT in L:
		        endT.append(bLabelend)
		self._cfg.append(bLabelend)        


                
	def visitWhileStat(self,ctx):
		if self._debug:
			print("while statement, condition is:")
			print(Trees.toStringTree(ctx.expr(),None,self._parser))
			print("and block is:")
			print(Trees.toStringTree(ctx.stat_block(),None,self._parser))
			
		labelbegin,labelend=self._cfg.newlabelWhile()
		bLabelbegin = BlockLABEL(labelbegin)

		#self._prog.addLabel(labelbegin)
		self._cfg.append(bLabelbegin)
		self.visit(ctx.expr())

		#self._prog.addInstructionBR("z",labelend)
		endCond = self._cfg.append(BlockBR("z",labelend))

		# List of statements
		self.visit(ctx.stat_block())

		#self._prog.addInstructionGOTO(labelbegin)
		#self._prog.addLabel(labelend)

		# End of while
		bgoto = self._cfg.append(BlockGOTO(labelbegin))
		bgoto.append(bLabelbegin)

		# Exit
		bLabelend = BlockLABEL(labelend)
		endCond.append(bLabelend)

		self._cfg.setEnd(bLabelend)

		return None

	def visitStatList(self,ctx):
		for stat in ctx.stat():
			#self._prog.addComment(Trees.toStringTree(stat,None,self._parser))
			#self._cfg.append(BlockComment(Trees.toStringTree(stat,None,self._parser)))
			self.visit(stat)

